import 'dotenv/config';
import { cac } from 'cac';
import consola from 'consola';
import { APIv1, APIv2, APIVariable, APIIconStatic, APIIconVariable, APILicense, fetchAPI, fetchVariable, parsev1, parsev2, generateAxis, parseVariable, parseIcons, parseLicenses } from 'google-font-metadata';
import colors from 'picocolors';
import { intro, group, text, multiselect, select, cancel, confirm, outro } from '@clack/prompts';
import fs from 'fs-extra';
import stringify from 'json-stringify-pretty-compact';
import * as path from 'pathe';
import path__default from 'pathe';
import { generateFontFace } from '@fontsource-utils/generate';
import PQueue from 'p-queue';
import flatten from 'flat';
import got from 'got';

var version = "0.2.0-beta.10";

const makeFontDownloadPath = (fontDir, fontId, subset, weight, style, extension) => `${fontDir}/files/${fontId}-${subset}-${weight}-${style}.${extension}`;
const makeVariableFontDownloadPath = (fontDir, fontId, subset, axes, style) => `${fontDir}/files/${fontId}-${subset}-${axes.toLowerCase()}-${style}.woff2`;
const makeFontFilePath = (fontId, subset, weight, style, extension) => `./files/${fontId}-${subset}-${weight}-${style}.${extension}`;
const makeVariableFontFilePath = (fontId, subset, axes, style) => `./files/${fontId}-${subset}-${axes}-${style}.woff2`;
const findClosest = (arr, num) => {
  const indexArr = arr.map((weight) => Math.abs(Number(weight) - num));
  const min = Math.min(...indexArr);
  const closest = arr[indexArr.indexOf(min)];
  return closest;
};
const licenseMap = {
  "apache license, version 2.0": "Apache-2.0",
  "sil open font license, 1.1": "OFL-1.1",
  "ubuntu font license, 1.0": "UFL-1.0",
  "mit license": "MIT"
};
const licenseShort = (license) => licenseMap[license.toLowerCase()];
const sassVar = (key, value) => `$${key}: ${value} !default;
`;

const axesValue = (metadata) => {
  let out = "(\n";
  for (const variable of Object.keys(metadata.variable)) {
    const axis = metadata.variable[variable];
    out += `  ${variable}: (
    default: ${axis.default},
    min: ${axis.min},
    max: ${axis.max},
    step: ${axis.step},
  ),
`;
  }
  out += ")";
  return out;
};
const unicodeValue = (unicode) => {
  if (Object.keys(unicode).length > 0) {
    let out = "(\n";
    for (const unicodeKey of Object.keys(unicode)) {
      const unicodeVal = unicode[unicodeKey];
      out += `  ${unicodeKey.replace("[", "").replace("]", "")}: (${unicodeVal}),
`;
    }
    out += ")";
    return out;
  }
  return "null";
};
const defaultAxis = (metadata) => {
  const axes = Object.keys(metadata.variable);
  if (axes.includes("wght"))
    return "wght";
  if (axes.includes("full"))
    return "full";
  return axes[0];
};
const sassMetadata = (metadata, unicode, isVariable) => {
  let out = "";
  out += sassVar("id", `'${metadata.id}'`);
  out += sassVar("family", `'${metadata.family}'`);
  out += sassVar("category", `${metadata.category}`);
  out += sassVar("subsets", `(${metadata.subsets.join(", ")})`);
  out += sassVar("weights", `(${metadata.weights.join(", ")})`);
  out += sassVar("styles", `(${metadata.styles.join(", ")})`);
  out += sassVar("axes", isVariable ? axesValue(metadata) : "null");
  out += sassVar(
    "defaults",
    `(
  subset: ${metadata.defSubset},
  weight: ${findClosest(metadata.weights, 400)},
  style: normal,
  axis: ${isVariable ? defaultAxis(metadata) : "null"},
)`
  );
  out += sassVar("unicode", unicodeValue(unicode));
  return out;
};

const sassMixins = `@use 'sass:list';
@use 'sass:map';
@use 'sass:math';
@use 'sass:meta';
@use 'sass:string';

@use 'metadata';

$metadata: meta.module-variables(metadata) !default;
$directory: null !default;

$family: null !default;
$display: null !default;
$displayVar: null !default;
$formats: null !default;
$subsets: null !default;
$weights: null !default;
$styles: null !default;
$axes: null !default;

@mixin generator(
  $metadata: $metadata,
  $directory: $directory,
  $family: $family,
  $display: $display,
  $displayVar: $displayVar,
  $formats: $formats,
  $subsets: $subsets,
  $weights: $weights,
  $styles: $styles,
  $axes: $axes
) {
  $directory: if(
    $directory,
    $directory,
    '~@fontsource#{if(map.get($metadata, axes), '-variable', '')}/#{map.get($metadata, id)}/files'
  );

  $family: if($family, $family, map.get($metadata, family) + if(map.get($metadata, axes), ' Variable', ''));
  $display: if($display, $display, swap);
  $displayVar: if($displayVar != null, $displayVar, true);
  $formats: if(not $formats or $formats == all, (woff2, woff), $formats);
  $subsets: if(
    $subsets,
    if($subsets == all, map.get($metadata, subsets), $subsets),
    map.get($metadata, defaults, subset)
  );
  $weights: if(
    $weights,
    if($weights == all, map.get($metadata, weights), $weights),
    map.get($metadata, defaults, weight)
  );
  $styles: if(
    $styles,
    if($styles == all, map.get($metadata, styles), $styles),
    map.get($metadata, defaults, style)
  );
  $axes: if(
    $axes,
    if($axes == all, full, $axes),
    if(
      map.get($metadata, axes),
      if(map.has-key($metadata, axes, wght), wght, full),
      null
    )
  );

  @each $subset in $subsets {
    @each $unicodeSubset, $unicodeRange in map.get($metadata, unicode) {
      @if (
        $subset ==
          $unicodeSubset or
          (
            // Is numeric subset
            $subset ==
              map.get($metadata, defaults, subset) and not
              map.has-key($metadata, subsets, $unicodeSubset)
          )
      ) {
        @each $weight in if($axes, null, $weights) {
          @each $axis in $axes {
            @each $style in $styles {
              $variant: '#{map.get($metadata, id)}-#{$unicodeSubset}-#{if($axis, $axis, $weight)}-#{$style}';

              $src: ();
              @each $format in $formats {
                $src: append(
                  $src,
                  url('#{$directory}/#{$variant}.#{$format}')
                    format('#{$format}#{if($axis, '-variations', '')}'),
                  comma
                );
              }

              @content ((
                metadata: $metadata,
                directory: $directory,
                family: $family,
                display: $display,
                displayVar: $displayVar,
                formats: $formats,
                subsets: $subsets,
                weights: $weights,
                styles: $styles,
                axes: $axes,

                variant: $variant,
                subset: $subset,
                unicodeSubset: $unicodeSubset,
                unicodeRange: $unicodeRange,
                weight: $weight,
                axis: $axis,
                style: $style,

                font-family: string.quote($family),
                font-style: if(
                  ($axis == full or $axis == slnt) and map.has-key($metadata, axes, slnt), 
                  oblique map.get($metadata, axes, slnt, min) + deg map.get($metadata, axes, slnt, max) + deg,
                  $style
                ),
                font-display: if($displayVar, var(--fontsource-display, $display), $display),
                font-weight: if(
                  ($axis == full or $axis == wght) and map.has-key($metadata, axes, wght), 
                  map.get($metadata, axes, wght, min) map.get($metadata, axes, wght, max),
                  $weight
                ),
                font-stretch: if(
                  ($axis == full or $axis == wdth) and map.has-key($metadata, axes, wdth), 
                  '#{map.get($metadata, axes, wdth, min)}% #{map.get($metadata, axes, wdth, max)}%',
                  null
                ),
                src: $src,
                unicode-range: $unicodeRange,
              ));
            }
          }
        }
      }
    }
  }
}

@mixin faces(
  $metadata: $metadata,
  $directory: $directory,
  $family: $family,
  $display: $display,
  $displayVar: $displayVar,
  $formats: $formats,
  $subsets: $subsets,
  $weights: $weights,
  $styles: $styles,
  $axes: $axes
) {
  @include generator(
      $metadata: $metadata,
      $directory: $directory,
      $family: $family,
      $display: $display,
      $displayVar: $displayVar,
      $formats: $formats,
      $subsets: $subsets,
      $weights: $weights,
      $styles: $styles,
      $axes: $axes
    )
    using ($data) {
    /* #{map.get($data, variant)} */
    @font-face {
      font-family: map.get($data, font-family);
      font-style: map.get($data, font-style);
      font-display: map.get($data, font-display);
      font-weight: map.get($data, font-weight);
      font-stretch: map.get($data, font-stretch);
      unicode-range: map.get($data, unicode-range);
      src: map.get($data, src);

      @content();
    }
  }
}
`;

const changelog = `# Changelog
See the [Fontsource CHANGELOG.md](https://github.com/fontsource/fontsource/blob/main/CHANGELOG.md).`;

const template = ({ id, family, license, type }, isVariable, oldVersion) => ({
  name: isVariable ? `@fontsource-variable/${id}` : `@fontsource/${id}`,
  version: oldVersion ?? JSON.parse(fs.readFileSync(path.join(process.cwd(), "package.json"), "utf8")).version,
  description: `Self-host the ${family} font in a neatly bundled NPM package.`,
  main: "index.css",
  publishConfig: { access: "public" },
  keywords: [
    "fontsource",
    "font",
    "font family",
    "google fonts",
    id,
    family,
    "css",
    "sass",
    "front-end",
    "web",
    "typeface",
    "variable"
  ],
  author: type === "other" ? license.attribution : "Google Inc.",
  license: license.type,
  homepage: `https://fontsource.org/fonts/${id}`,
  repository: {
    type: "git",
    url: "https://github.com/fontsource/font-files.git",
    directory: `fonts/${isVariable ? "variable" : type}/${id}`
  }
});
const packageJson = async (metadata, opts) => {
  let oldVersion;
  try {
    await fs.access(path.join(opts.dir, "package.json"));
    oldVersion = JSON.parse(
      await fs.readFile(path.join(opts.dir, "package.json"), "utf8")
    ).version;
  } catch {
  }
  const file = template(metadata, opts.isVariable, oldVersion);
  await fs.writeFile(path.join(opts.dir, "package.json"), stringify(file));
};

const variableText = () => `## Variable Fonts

This particular typeface supports [variable fonts](https://developer.mozilla.org/en-US/docs/Web/CSS/CSS_Fonts/Variable_Fonts_Guide).
Variable documentation can be found [here](https://fontsource.org/docs/variable-fonts).

`;
const readme = ({
  id,
  family,
  version,
  weights,
  styles,
  subsets,
  variable,
  license
}, isVariable) => `# Fontsource ${family}

[![npm (scoped)](https://img.shields.io/npm/v/@fontsource/${id}?color=brightgreen)](https://www.npmjs.com/package/@fontsource/${id}) [![Generic badge](https://img.shields.io/badge/fontsource-passing-brightgreen)](https://github.com/fontsource/fontsource) [![Monthly downloads](https://badgen.net/npm/dm/@fontsource/${id})](https://github.com/fontsource/fontsource) [![Total downloads](https://badgen.net/npm/dt/@fontsource/${id})](https://github.com/fontsource/fontsource) [![GitHub stars](https://img.shields.io/github/stars/fontsource/fontsource.svg?style=social&label=Star)](https://github.com/fontsource/fontsource/stargazers)

The CSS and web font files to easily self-host the \u201C${family}\u201D font. Please visit the main [Fontsource website](https://fontsource.org/fonts/${id}) to view more details on this package.

## Quick Installation

Fontsource has a variety of methods to import CSS, such as using a bundler like Webpack. Alternatively, it supports SASS. Full documentation can be found [here](https://fontsource.org/docs/introduction).

\`\`\`javascript
npm install ${isVariable ? `@fontsource-variable/${id}` : `@fontsource/${id}`}
\`\`\`

Within your app entry file or site component, import it in.

\`\`\`javascript
${isVariable ? `import "@fontsource-variable/${id}"; // Defaults to wght axis
import "@fontsource-variable/${id}/wght.css"; // Specify axis
import "@fontsource-variable/${id}/wght-italic.css"; // Specify axis and style` : `import "@fontsource/${id}"; // Defaults to weight 400
import "@fontsource/${id}/400.css"; // Specify weight
import "@fontsource/${id}/400-italic.css"; // Specify weight and style`}

\`\`\`

Supported variables:
- Weights: \`[${weights}]\`
- Styles: \`[${styles}]\`
- Subsets: \`[${subsets}]\`${isVariable ? `
- Axes: \`[${Object.keys(variable).filter((axis) => axis !== "ital")}]\`` : ""}

Finally, you can reference the font name in a CSS stylesheet, CSS Module, or CSS-in-JS.

\`\`\`css
body {
  font-family: "${family};
}
\`\`\`

${variable ? variableText() : ""}## Licensing
It is important to always read the license for every font that you use.
Most of the fonts in the collection use the SIL Open Font License, v1.1. Some fonts use the Apache 2 license. The Ubuntu fonts use the Ubuntu Font License v1.0.

${license.attribution}
[${license.type}](${license.url})

## Other Notes
Font version (provided by source): \`${version}\`.
Feel free to star and contribute new ideas to this repository that aim to improve the performance of font loading, as well as expanding the existing library we already have. Any suggestions or ideas can be voiced via an [issue](https://github.com/fontsource/fontsource/issues).`;

const packagerCustom = async (metadata) => {
  const { id, family, subsets, weights, styles } = metadata;
  const dir = `./${id}`;
  const indexWeight = findClosest(weights, 400);
  for (const subset of subsets) {
    const cssSubset = [];
    const cssSubsetItalic = [];
    for (const weight of weights) {
      for (const style of styles) {
        const fontObj = {
          family,
          style,
          display: "swap",
          weight,
          src: [
            {
              url: makeFontFilePath(id, subset, weight, style, "woff2"),
              format: "woff2"
            },
            {
              url: makeFontFilePath(id, subset, weight, style, "woff"),
              format: "woff"
            }
          ],
          comment: `${id}-${subset}-${weight}-${style} CUSTOM`,
          displayVar: true
        };
        const css = generateFontFace(fontObj);
        if (style === "normal") {
          await fs.writeFile(path__default.join(dir, `${subset}-${weight}.css`), css);
          cssSubset.push(css);
        } else {
          await fs.writeFile(path__default.join(dir, `${subset}-${weight}-${style}.css`), css);
          cssSubsetItalic.push(css);
        }
      }
      if (weight === indexWeight) {
        await fs.writeFile(path__default.join(dir, "index.css"), cssSubset.join("\n\n"));
      }
      await fs.writeFile(path__default.join(dir, `${subset}.css`), cssSubset.join("\n\n"));
      if (cssSubsetItalic.length > 0) {
        await fs.writeFile(path__default.join(dir, `${subset}-italic.css`), cssSubsetItalic.join("\n\n"));
      }
    }
  }
};

const buildCustom = async (metadata) => {
  const dir = `./${metadata.id}`;
  fs.ensureDirSync(dir);
  fs.ensureDirSync(path__default.join(dir, "/files"));
  await packagerCustom(metadata);
  await fs.mkdir(path__default.join(dir, "scss"));
  await fs.writeFile(
    path__default.join(dir, "scss/metadata.scss"),
    sassMetadata(metadata, {}, false)
  );
  await fs.writeFile(path__default.join(dir, "scss/mixins.scss"), sassMixins);
  await fs.writeFile(path__default.join(dir, "README.md"), readme(metadata, false));
  await fs.writeFile(path__default.join(dir, "metadata.json"), stringify(metadata));
  await fs.writeFile(path__default.join(dir, "unicode.json"), stringify({}));
  await fs.writeFile(path__default.join(dir, "CHANGELOG.md"), changelog);
  await packageJson(metadata, {
    dir,
    isVariable: false,
    tmpDir: "",
    force: false
  });
};

const create = async () => {
  intro(colors.cyan(colors.bold("fontsource")));
  const cfg = await group(
    {
      name: () => text({
        message: colors.bold("What is the name of the font?"),
        placeholder: "Noto Sans JP",
        validate(value) {
          if (!value)
            return "Please enter a name";
        }
      }),
      subsets: () => text({
        message: colors.bold("What are the subsets of the font? (separate by commas)"),
        placeholder: "latin, latin-ext, japanese",
        validate(value) {
          if (!value)
            return "Please enter at least one subset";
        }
      }),
      weights: () => text({
        message: colors.bold("What are the weights of the font? (separate by commas)"),
        placeholder: "100, 200, 300, 400, 500, 600, 700, 800, 900",
        validate(value) {
          if (!value)
            return "Please enter at least one weight";
          const weights = value.split(",").map(Number);
          if (weights.some((weight) => Number.isNaN(weight)))
            return "Please enter only numbers";
        }
      }),
      styles: () => multiselect({
        message: colors.bold("What are the styles of the font?"),
        options: [
          { value: "normal", label: "Normal" },
          { value: "italic", label: "Italic" }
        ],
        required: true
      }),
      version: () => text({
        message: colors.bold("What is the version of the font?"),
        placeholder: "v1.0",
        validate(value) {
          if (!value)
            return "Please enter a version";
        }
      }),
      category: () => select({
        message: colors.bold("What is the category of the font?"),
        options: [
          { value: "sans-serif", label: "Sans Serif" },
          { value: "serif", label: "Serif" },
          { value: "display", label: "Display" },
          { value: "monospace", label: "Monospace" },
          { value: "handwriting", label: "Handwriting" },
          { value: "icons", label: "Icons" },
          { value: "other", label: "Other" }
        ]
      }),
      license: () => select({
        message: colors.bold("What is the license of the font?"),
        options: [
          { value: "sil open font license, 1.1", label: "SIL Open Font License (OFL-1.1)" },
          { value: "apache license, version 2.0", label: "Apache License 2.0 (Apache-2.0)" },
          { value: "ubuntu font license, 1.0", label: "Ubuntu Font License (UFL-1.0)" },
          { value: "mit", label: "MIT License" },
          { value: "other", label: "Other License", hint: "Please make an issue verifying if this is usable! We are likely to reject PRs that do not match the above licenses." }
        ]
      }),
      licenseUrl: () => text({
        message: colors.bold("What is the URL of the license?"),
        validate(value) {
          if (!value)
            return "Please enter a URL";
        }
      }),
      licenseAttribution: () => text({
        message: colors.bold("What is the author name or attribution of the font?"),
        validate(value) {
          if (!value)
            return "Please enter a name or attribution";
        }
      }),
      sourceUrl: () => text({
        message: colors.bold("What is the URL of the source files? GitHub repositories are preferred"),
        validate(value) {
          if (!value)
            return "Please enter a URL";
        }
      })
    },
    {
      onCancel: () => {
        cancel("Package creation cancelled");
        process.exit(0);
      }
    }
  );
  const metadata = {
    id: cfg.name.toLowerCase().replace(/ /g, "-"),
    family: cfg.name,
    subsets: cfg.subsets.split(",").map((subset) => subset.trim()),
    weights: cfg.weights.split(",").map((weight) => Number.parseInt(weight.trim(), 10)),
    styles: cfg.styles,
    defSubset: cfg.subsets.split(",")[0].trim(),
    variable: false,
    lastModified: (/* @__PURE__ */ new Date()).toISOString(),
    version: cfg.version,
    category: cfg.category,
    license: {
      type: licenseShort(cfg.license) ?? cfg.license,
      url: cfg.licenseUrl,
      attribution: cfg.licenseAttribution
    },
    source: cfg.sourceUrl,
    type: "other"
  };
  const confirmCreate = await confirm({
    message: colors.bold(`Create package in current directory? ./${metadata.id}`)
  });
  if (!confirmCreate) {
    cancel("Package creation cancelled.");
    process.exit(0);
  }
  try {
    await buildCustom(metadata);
  } catch (error) {
    cancel("Package creation failed.");
    consola.error(error);
    return;
  }
  outro(colors.green(`You're all set!

	Now copy all the font files into the files directory and run "fontsource create-verify" to verify your package.

	Please ensure the file names match the format "${metadata.id}-subset-weight-style.extension"
	Example: "${metadata.id}-latin-400-normal.woff2" or "${metadata.id}-latin-ext-700-italic.woff"

	Please also copy the appropriate LICENSE file from your license URL to the root directory of the package.`));
};

const verifyFilenames = async (metadata, dir) => {
  const filenames = await fs.readdir(path__default.join(dir, "files"));
  if (filenames.length === 0) {
    throw new Error("No files found in the files directory.");
  }
  const expectedFilenames = [];
  for (const subset of metadata.subsets) {
    for (const weight of metadata.weights) {
      for (const style of metadata.styles) {
        expectedFilenames.push(`${metadata.id}-${subset}-${weight}-${style}.woff2`, `${metadata.id}-${subset}-${weight}-${style}.woff`);
      }
    }
  }
  const missingFilenames = expectedFilenames.filter((filename) => !filenames.includes(filename));
  if (missingFilenames.length > 0) {
    throw new Error(`The following files are missing: ${missingFilenames.join(", ")}

	Please ensure the file names match the format "${metadata.id}-subset-weight-style.extension"
	Example: "${metadata.id}-latin-400-normal.woff2" or "${metadata.id}-latin-ext-700-italic.woff"`);
  }
  const nonMatchingFilenames = filenames.filter((filename) => !expectedFilenames.includes(filename));
  if (nonMatchingFilenames.length > 0) {
    throw new Error(`The following files are not expected: ${nonMatchingFilenames.join(", ")}

	Please ensure the file names match the format "${metadata.id}-subset-weight-style.extension"
	Example: "${metadata.id}-latin-400-normal.woff2" or "${metadata.id}-latin-ext-700-italic.woff"`);
  }
};
const verify = async () => {
  intro(colors.cyan(colors.bold("fontsource")));
  const cfg = await group({
    id: () => text({
      message: colors.bold("What is the ID of the font?"),
      placeholder: "noto-sans-jp",
      validate(value) {
        if (!value)
          return "Please enter an ID.";
      }
    })
  });
  if (!await fs.pathExists(`./${cfg.id}/files`)) {
    cancel("The directory does not exist.");
    return;
  }
  if (!await fs.pathExists(`./${cfg.id}/metadata.json`)) {
    cancel("The metadata.json file does not exist.");
    return;
  }
  if (!await fs.pathExists(`./${cfg.id}/LICENSE`)) {
    cancel("The LICENSE file does not exist.");
    return;
  }
  try {
    const dir = `./${cfg.id}`;
    const metadata = await fs.readJson(path__default.join(dir, "metadata.json"));
    await verifyFilenames(metadata, dir);
  } catch (error) {
    cancel(error.message);
    return;
  }
  outro(colors.green("All checks passed! Feel free to send a PR over to the main repo adding the package to the appropriate fonts directory."));
};

const gotDownload = async (url, dest) => {
  const response = await got(url).buffer();
  await fs.writeFile(dest, response);
};
const pairGenerator = (variants) => {
  const flattenedPairs = Object.entries(flatten(variants));
  const splitPairs = flattenedPairs.map((pair) => [
    pair[0].split("."),
    pair[1]
  ]);
  const ABSOLUTE_URL_REGEX = /^https?:\/\//;
  const urlPairs = splitPairs.filter(
    (pair) => ABSOLUTE_URL_REGEX.test(pair[1].toString())
  );
  const cleanedPairs = urlPairs.filter((pair) => {
    const extension = pair[0][4];
    if (extension === "truetype" || extension === "opentype") {
      return false;
    }
    return true;
  });
  return cleanedPairs;
};
const generateLinks = (id, opts) => {
  const fontV1 = APIv1[id];
  const fontV2 = APIv2[id];
  let downloadURLPairsV1 = pairGenerator(fontV1.variants);
  const downloadURLPairsV2 = pairGenerator(fontV2.variants);
  let hasUnicodeSubsets = false;
  const re = /\[.*?]/g;
  for (const pair of downloadURLPairsV2) {
    if (re.test(pair[0][2])) {
      hasUnicodeSubsets = true;
    }
  }
  if (!hasUnicodeSubsets) {
    downloadURLPairsV1 = downloadURLPairsV1.filter(
      (pair) => pair[0][4] !== "woff2"
    );
  }
  const linksV1 = downloadURLPairsV1.map((pair) => {
    const types = pair[0];
    const dest = makeFontDownloadPath(
      opts.dir,
      id,
      types[2],
      Number(types[0]),
      types[1],
      types[4]
    );
    const url = pair[1];
    return {
      url,
      dest
    };
  });
  const linksV2 = downloadURLPairsV2.map((pair) => {
    const types = pair[0];
    const dest = makeFontDownloadPath(
      opts.dir,
      id,
      types[2].replace("[", "").replace("]", ""),
      Number(types[0]),
      types[1],
      types[4]
    );
    const url = pair[1];
    return { url, dest };
  });
  const links = [...linksV1, ...linksV2];
  return links;
};
const variableLinks = (id, opts) => {
  const fontVariable = APIVariable[id];
  const downloadURLPairsVariable = pairGenerator(fontVariable.variants);
  const links = downloadURLPairsVariable.map((pair) => {
    const types = pair[0];
    const dest = makeVariableFontDownloadPath(
      opts.dir,
      id,
      types[2].replace("[", "").replace("]", ""),
      types[0],
      types[1]
    );
    const url = pair[1];
    return {
      url,
      dest
    };
  });
  return links;
};
const iconStaticLinks = (id, opts) => {
  const iconStatic = APIIconStatic[id];
  const downloadURLPairsIconStatic = pairGenerator(iconStatic.variants);
  const links = downloadURLPairsIconStatic.map((pair) => {
    const types = pair[0];
    const dest = makeFontDownloadPath(
      opts.dir,
      id,
      types[2].replace("[", "").replace("]", ""),
      Number(types[0]),
      types[1],
      types[4]
    );
    const url = pair[1];
    return {
      url,
      dest
    };
  });
  return links;
};
const iconVariableLinks = (id, opts) => {
  const iconVariable = APIIconVariable[id];
  const downloadURLPairsIconVariable = pairGenerator(iconVariable.variants);
  const links = downloadURLPairsIconVariable.map((pair) => {
    const types = pair[0];
    const dest = makeVariableFontDownloadPath(
      opts.dir,
      id,
      types[2].replace("[", "").replace("]", ""),
      types[0],
      types[1]
    );
    const url = pair[1];
    return { url, dest };
  });
  return links;
};
const queue$1 = new PQueue({ concurrency: 60 });
const download = async (id, opts) => {
  await fs.ensureDir(path.join(opts.dir, "files"));
  let links;
  if (opts.isIcon) {
    links = opts.isVariable ? iconVariableLinks(id, opts) : iconStaticLinks(id, opts);
  } else {
    links = opts.isVariable ? variableLinks(id, opts) : generateLinks(id, opts);
  }
  for (const link of links) {
    queue$1.add(() => gotDownload(link.url, link.dest));
  }
  await queue$1.onIdle();
};

const apache = (copyright) => `                                 Apache License
                           Version 2.0, January 2004
                        http://www.apache.org/licenses/

   TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION

   1. Definitions.

      "License" shall mean the terms and conditions for use, reproduction,
      and distribution as defined by Sections 1 through 9 of this document.

      "Licensor" shall mean the copyright owner or entity authorized by
      the copyright owner that is granting the License.

      "Legal Entity" shall mean the union of the acting entity and all
      other entities that control, are controlled by, or are under common
      control with that entity. For the purposes of this definition,
      "control" means (i) the power, direct or indirect, to cause the
      direction or management of such entity, whether by contract or
      otherwise, or (ii) ownership of fifty percent (50%) or more of the
      outstanding shares, or (iii) beneficial ownership of such entity.

      "You" (or "Your") shall mean an individual or Legal Entity
      exercising permissions granted by this License.

      "Source" form shall mean the preferred form for making modifications,
      including but not limited to software source code, documentation
      source, and configuration files.

      "Object" form shall mean any form resulting from mechanical
      transformation or translation of a Source form, including but
      not limited to compiled object code, generated documentation,
      and conversions to other media types.

      "Work" shall mean the work of authorship, whether in Source or
      Object form, made available under the License, as indicated by a
      copyright notice that is included in or attached to the work
      (an example is provided in the Appendix below).

      "Derivative Works" shall mean any work, whether in Source or Object
      form, that is based on (or derived from) the Work and for which the
      editorial revisions, annotations, elaborations, or other modifications
      represent, as a whole, an original work of authorship. For the purposes
      of this License, Derivative Works shall not include works that remain
      separable from, or merely link (or bind by name) to the interfaces of,
      the Work and Derivative Works thereof.

      "Contribution" shall mean any work of authorship, including
      the original version of the Work and any modifications or additions
      to that Work or Derivative Works thereof, that is intentionally
      submitted to Licensor for inclusion in the Work by the copyright owner
      or by an individual or Legal Entity authorized to submit on behalf of
      the copyright owner. For the purposes of this definition, "submitted"
      means any form of electronic, verbal, or written communication sent
      to the Licensor or its representatives, including but not limited to
      communication on electronic mailing lists, source code control systems,
      and issue tracking systems that are managed by, or on behalf of, the
      Licensor for the purpose of discussing and improving the Work, but
      excluding communication that is conspicuously marked or otherwise
      designated in writing by the copyright owner as "Not a Contribution."

      "Contributor" shall mean Licensor and any individual or Legal Entity
      on behalf of whom a Contribution has been received by Licensor and
      subsequently incorporated within the Work.

   2. Grant of Copyright License. Subject to the terms and conditions of
      this License, each Contributor hereby grants to You a perpetual,
      worldwide, non-exclusive, no-charge, royalty-free, irrevocable
      copyright license to reproduce, prepare Derivative Works of,
      publicly display, publicly perform, sublicense, and distribute the
      Work and such Derivative Works in Source or Object form.

   3. Grant of Patent License. Subject to the terms and conditions of
      this License, each Contributor hereby grants to You a perpetual,
      worldwide, non-exclusive, no-charge, royalty-free, irrevocable
      (except as stated in this section) patent license to make, have made,
      use, offer to sell, sell, import, and otherwise transfer the Work,
      where such license applies only to those patent claims licensable
      by such Contributor that are necessarily infringed by their
      Contribution(s) alone or by combination of their Contribution(s)
      with the Work to which such Contribution(s) was submitted. If You
      institute patent litigation against any entity (including a
      cross-claim or counterclaim in a lawsuit) alleging that the Work
      or a Contribution incorporated within the Work constitutes direct
      or contributory patent infringement, then any patent licenses
      granted to You under this License for that Work shall terminate
      as of the date such litigation is filed.

   4. Redistribution. You may reproduce and distribute copies of the
      Work or Derivative Works thereof in any medium, with or without
      modifications, and in Source or Object form, provided that You
      meet the following conditions:

      (a) You must give any other recipients of the Work or
          Derivative Works a copy of this License; and

      (b) You must cause any modified files to carry prominent notices
          stating that You changed the files; and

      (c) You must retain, in the Source form of any Derivative Works
          that You distribute, all copyright, patent, trademark, and
          attribution notices from the Source form of the Work,
          excluding those notices that do not pertain to any part of
          the Derivative Works; and

      (d) If the Work includes a "NOTICE" text file as part of its
          distribution, then any Derivative Works that You distribute must
          include a readable copy of the attribution notices contained
          within such NOTICE file, excluding those notices that do not
          pertain to any part of the Derivative Works, in at least one
          of the following places: within a NOTICE text file distributed
          as part of the Derivative Works; within the Source form or
          documentation, if provided along with the Derivative Works; or,
          within a display generated by the Derivative Works, if and
          wherever such third-party notices normally appear. The contents
          of the NOTICE file are for informational purposes only and
          do not modify the License. You may add Your own attribution
          notices within Derivative Works that You distribute, alongside
          or as an addendum to the NOTICE text from the Work, provided
          that such additional attribution notices cannot be construed
          as modifying the License.

      You may add Your own copyright statement to Your modifications and
      may provide additional or different license terms and conditions
      for use, reproduction, or distribution of Your modifications, or
      for any such Derivative Works as a whole, provided Your use,
      reproduction, and distribution of the Work otherwise complies with
      the conditions stated in this License.

   5. Submission of Contributions. Unless You explicitly state otherwise,
      any Contribution intentionally submitted for inclusion in the Work
      by You to the Licensor shall be under the terms and conditions of
      this License, without any additional terms or conditions.
      Notwithstanding the above, nothing herein shall supersede or modify
      the terms of any separate license agreement you may have executed
      with Licensor regarding such Contributions.

   6. Trademarks. This License does not grant permission to use the trade
      names, trademarks, service marks, or product names of the Licensor,
      except as required for reasonable and customary use in describing the
      origin of the Work and reproducing the content of the NOTICE file.

   7. Disclaimer of Warranty. Unless required by applicable law or
      agreed to in writing, Licensor provides the Work (and each
      Contributor provides its Contributions) on an "AS IS" BASIS,
      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
      implied, including, without limitation, any warranties or conditions
      of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
      PARTICULAR PURPOSE. You are solely responsible for determining the
      appropriateness of using or redistributing the Work and assume any
      risks associated with Your exercise of permissions under this License.

   8. Limitation of Liability. In no event and under no legal theory,
      whether in tort (including negligence), contract, or otherwise,
      unless required by applicable law (such as deliberate and grossly
      negligent acts) or agreed to in writing, shall any Contributor be
      liable to You for damages, including any direct, indirect, special,
      incidental, or consequential damages of any character arising as a
      result of this License or out of the use or inability to use the
      Work (including but not limited to damages for loss of goodwill,
      work stoppage, computer failure or malfunction, or any and all
      other commercial damages or losses), even if such Contributor
      has been advised of the possibility of such damages.

   9. Accepting Warranty or Additional Liability. While redistributing
      the Work or Derivative Works thereof, You may choose to offer,
      and charge a fee for, acceptance of support, warranty, indemnity,
      or other liability obligations and/or rights consistent with this
      License. However, in accepting such obligations, You may act only
      on Your own behalf and on Your sole responsibility, not on behalf
      of any other Contributor, and only if You agree to indemnify,
      defend, and hold each Contributor harmless for any liability
      incurred by, or claims asserted against, such Contributor by reason
      of your accepting any such warranty or additional liability.

   END OF TERMS AND CONDITIONS

   APPENDIX: How to apply the Apache License to your work.

      To apply the Apache License to your work, attach the following
      boilerplate notice, with the fields enclosed by brackets "[]"
      replaced with your own identifying information. (Don't include
      the brackets!)  The text should be enclosed in the appropriate
      comment syntax for the file format. We also recommend that a
      file or class name and description of purpose be included on the
      same "printed page" as the copyright notice for easier
      identification within third-party archives.

   ${copyright}

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.`;

const ofl = (copyright) => `${copyright}

This Font Software is licensed under the SIL Open Font License, Version 1.1.
This license is copied below, and is also available with a FAQ at:
http://scripts.sil.org/OFL


-----------------------------------------------------------
SIL OPEN FONT LICENSE Version 1.1 - 26 February 2007
-----------------------------------------------------------

PREAMBLE
The goals of the Open Font License (OFL) are to stimulate worldwide
development of collaborative font projects, to support the font creation
efforts of academic and linguistic communities, and to provide a free and
open framework in which fonts may be shared and improved in partnership
with others.

The OFL allows the licensed fonts to be used, studied, modified and
redistributed freely as long as they are not sold by themselves. The
fonts, including any derivative works, can be bundled, embedded,
redistributed and/or sold with any software provided that any reserved
names are not used by derivative works. The fonts and derivatives,
however, cannot be released under any other type of license. The
requirement for fonts to remain under this license does not apply
to any document created using the fonts or their derivatives.

DEFINITIONS
"Font Software" refers to the set of files released by the Copyright
Holder(s) under this license and clearly marked as such. This may
include source files, build scripts and documentation.

"Reserved Font Name" refers to any names specified as such after the
copyright statement(s).

"Original Version" refers to the collection of Font Software components as
distributed by the Copyright Holder(s).

"Modified Version" refers to any derivative made by adding to, deleting,
or substituting -- in part or in whole -- any of the components of the
Original Version, by changing formats or by porting the Font Software to a
new environment.

"Author" refers to any designer, engineer, programmer, technical
writer or other person who contributed to the Font Software.

PERMISSION & CONDITIONS
Permission is hereby granted, free of charge, to any person obtaining
a copy of the Font Software, to use, study, copy, merge, embed, modify,
redistribute, and sell modified and unmodified copies of the Font
Software, subject to the following conditions:

1) Neither the Font Software nor any of its individual components,
in Original or Modified Versions, may be sold by itself.

2) Original or Modified Versions of the Font Software may be bundled,
redistributed and/or sold with any software, provided that each copy
contains the above copyright notice and this license. These can be
included either as stand-alone text files, human-readable headers or
in the appropriate machine-readable metadata fields within text or
binary files as long as those fields can be easily viewed by the user.

3) No Modified Version of the Font Software may use the Reserved Font
Name(s) unless explicit written permission is granted by the corresponding
Copyright Holder. This restriction only applies to the primary font name as
presented to the users.

4) The name(s) of the Copyright Holder(s) or the Author(s) of the Font
Software shall not be used to promote, endorse or advertise any
Modified Version, except to acknowledge the contribution(s) of the
Copyright Holder(s) and the Author(s) or with their explicit written
permission.

5) The Font Software, modified or unmodified, in part or in whole,
must be distributed entirely under this license, and must not be
distributed under any other license. The requirement for fonts to
remain under this license does not apply to any document created
using the Font Software.

TERMINATION
This license becomes null and void if any of the above conditions are
not met.

DISCLAIMER
THE FONT SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO ANY WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT
OF COPYRIGHT, PATENT, TRADEMARK, OR OTHER RIGHT. IN NO EVENT SHALL THE
COPYRIGHT HOLDER BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
INCLUDING ANY GENERAL, SPECIAL, INDIRECT, INCIDENTAL, OR CONSEQUENTIAL
DAMAGES, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF THE USE OR INABILITY TO USE THE FONT SOFTWARE OR FROM
OTHER DEALINGS IN THE FONT SOFTWARE.
`;

const ufl = () => `-------------------------------
UBUNTU FONT LICENCE Version 1.0
-------------------------------

PREAMBLE
This licence allows the licensed fonts to be used, studied, modified and
redistributed freely. The fonts, including any derivative works, can be
bundled, embedded, and redistributed provided the terms of this licence
are met. The fonts and derivatives, however, cannot be released under
any other licence. The requirement for fonts to remain under this
licence does not require any document created using the fonts or their
derivatives to be published under this licence, as long as the primary
purpose of the document is not to be a vehicle for the distribution of
the fonts.

DEFINITIONS
"Font Software" refers to the set of files released by the Copyright
Holder(s) under this licence and clearly marked as such. This may
include source files, build scripts and documentation.

"Original Version" refers to the collection of Font Software components
as received under this licence.

"Modified Version" refers to any derivative made by adding to, deleting,
or substituting -- in part or in whole -- any of the components of the
Original Version, by changing formats or by porting the Font Software to
a new environment.

"Copyright Holder(s)" refers to all individuals and companies who have a
copyright ownership of the Font Software.

"Substantially Changed" refers to Modified Versions which can be easily
identified as dissimilar to the Font Software by users of the Font
Software comparing the Original Version with the Modified Version.

To "Propagate" a work means to do anything with it that, without
permission, would make you directly or secondarily liable for
infringement under applicable copyright law, except executing it on a
computer or modifying a private copy. Propagation includes copying,
distribution (with or without modification and with or without charging
a redistribution fee), making available to the public, and in some
countries other activities as well.

PERMISSION & CONDITIONS
This licence does not grant any rights under trademark law and all such
rights are reserved.

Permission is hereby granted, free of charge, to any person obtaining a
copy of the Font Software, to propagate the Font Software, subject to
the below conditions:

1) Each copy of the Font Software must contain the above copyright
notice and this licence. These can be included either as stand-alone
text files, human-readable headers or in the appropriate machine-
readable metadata fields within text or binary files as long as those
fields can be easily viewed by the user.

2) The font name complies with the following:
(a) The Original Version must retain its name, unmodified.
(b) Modified Versions which are Substantially Changed must be renamed to
avoid use of the name of the Original Version or similar names entirely.
(c) Modified Versions which are not Substantially Changed must be
renamed to both (i) retain the name of the Original Version and (ii) add
additional naming elements to distinguish the Modified Version from the
Original Version. The name of such Modified Versions must be the name of
the Original Version, with "derivative X" where X represents the name of
the new work, appended to that name.

3) The name(s) of the Copyright Holder(s) and any contributor to the
Font Software shall not be used to promote, endorse or advertise any
Modified Version, except (i) as required by this licence, (ii) to
acknowledge the contribution(s) of the Copyright Holder(s) or (iii) with
their explicit written permission.

4) The Font Software, modified or unmodified, in part or in whole, must
be distributed entirely under this licence, and must not be distributed
under any other licence. The requirement for fonts to remain under this
licence does not affect any document created using the Font Software,
except any version of the Font Software extracted from a document
created using the Font Software may only be distributed under this
licence.

TERMINATION
This licence becomes null and void if any of the above conditions are
not met.

DISCLAIMER
THE FONT SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO ANY WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT OF
COPYRIGHT, PATENT, TRADEMARK, OR OTHER RIGHT. IN NO EVENT SHALL THE
COPYRIGHT HOLDER BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
INCLUDING ANY GENERAL, SPECIAL, INDIRECT, INCIDENTAL, OR CONSEQUENTIAL
DAMAGES, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF THE USE OR INABILITY TO USE THE FONT SOFTWARE OR FROM OTHER
DEALINGS IN THE FONT SOFTWARE.`;

const generateLicense = async (id, licenseType, opts) => {
  const licensePath = path.join(opts.dir, "LICENSE");
  const copyright = APILicense[id].original;
  let txt;
  switch (licenseType.toLowerCase()) {
    case "apache license, version 2.0": {
      txt = apache(copyright);
      break;
    }
    case "sil open font license, 1.1": {
      txt = ofl(copyright);
      break;
    }
    case "ubuntu font license, 1.0": {
      txt = ufl();
      break;
    }
    default: {
      throw new Error(`Unknown license type: ${licenseType}`);
    }
  }
  await fs.writeFile(licensePath, txt);
};

const packagerV1 = async (id, opts) => {
  const { family, styles, weights, subsets, variants } = APIv1[id];
  for (const subset of subsets) {
    const cssSubset = [];
    const cssSubsetItalic = [];
    for (const weight of weights) {
      for (const style of styles) {
        if (style in variants[weight]) {
          const fontObj = {
            family,
            style,
            display: "swap",
            weight,
            src: [
              {
                url: makeFontFilePath(id, subset, weight, style, "woff2"),
                format: "woff2"
              },
              {
                url: makeFontFilePath(id, subset, weight, style, "woff"),
                format: "woff"
              }
            ],
            comment: `${id}-${subset}-${weight}-${style}`,
            displayVar: true
          };
          const css = generateFontFace(fontObj);
          if (style === "normal") {
            const cssPath = path.join(opts.dir, `${subset}-${weight}.css`);
            await fs.writeFile(cssPath, css);
            cssSubset.push(css);
          } else {
            const cssStylePath = path.join(
              opts.dir,
              `${subset}-${weight}-${style}.css`
            );
            await fs.writeFile(cssStylePath, css);
            cssSubsetItalic.push(css);
          }
        }
      }
    }
    const cssSubsetPath = path.join(opts.dir, `${subset}.css`);
    await fs.writeFile(cssSubsetPath, cssSubset.join("\n\n"));
    if (cssSubsetItalic.length > 0) {
      const cssSubsetItalicPath = path.join(opts.dir, `${subset}-italic.css`);
      await fs.writeFile(cssSubsetItalicPath, cssSubsetItalic.join("\n\n"));
    }
  }
};

const packagerV2 = async (id, opts) => {
  const { family, styles, weights, variants, unicodeRange } = APIv2[id];
  const indexWeight = findClosest(weights, 400);
  const unicodeKeys = Object.keys(unicodeRange);
  for (const weight of weights) {
    for (const style of styles) {
      const cssStyle = [];
      for (const subset of unicodeKeys) {
        const cleanedSubset = subset.replace("[", "").replace("]", "");
        if (style in variants[weight]) {
          const fontObj = {
            family,
            style,
            display: "swap",
            weight,
            unicodeRange: unicodeRange[subset],
            src: [
              {
                url: makeFontFilePath(
                  id,
                  cleanedSubset,
                  weight,
                  style,
                  "woff2"
                ),
                format: "woff2"
              },
              {
                url: makeFontFilePath(id, cleanedSubset, weight, style, "woff"),
                format: "woff"
              }
            ],
            comment: `${id}-${subset}-${weight}-${style}`,
            displayVar: true
          };
          const css = generateFontFace(fontObj);
          cssStyle.push(css);
        }
      }
      if (style in variants[weight]) {
        if (style === "normal") {
          const cssPath = path.join(opts.dir, `${weight}.css`);
          await fs.writeFile(cssPath, cssStyle.join("\n\n"));
          if (weight === indexWeight) {
            await fs.writeFile(
              path.join(opts.dir, "index.css"),
              cssStyle.join("\n\n")
            );
          }
        } else {
          const cssStylePath = path.join(opts.dir, `${weight}-${style}.css`);
          await fs.writeFile(cssStylePath, cssStyle.join("\n\n"));
        }
      }
    }
  }
};

const packagerVariable = async (id, opts) => {
  const font = APIv2[id];
  const fontVariable = APIVariable[id];
  let indexCSS = "";
  for (const axes of Object.keys(fontVariable.variants)) {
    const variant = fontVariable.variants[axes];
    const styles = Object.keys(variant);
    const axesLower = axes.toLowerCase();
    const variableOpts = {
      wght: fontVariable.axes.wght
    };
    if (axes === "standard" || axes === "full" || axes === "wdth")
      variableOpts.stretch = fontVariable.axes.wdth;
    if (axes === "standard" || axes === "full" || axes === "slnt")
      variableOpts.slnt = fontVariable.axes.slnt;
    for (const style of styles) {
      const cssStyle = [];
      for (const subset of Object.keys(variant[style])) {
        const fontObj = {
          family: font.family,
          style,
          display: "swap",
          weight: findClosest(font.weights, 400),
          unicodeRange: font.unicodeRange[subset],
          variable: variableOpts,
          src: [
            {
              url: makeVariableFontFilePath(id, subset, axesLower, style),
              format: "woff2-variations"
            }
          ],
          comment: `${id}-${subset}-${axesLower}-${style}`,
          displayVar: true
        };
        const css2 = generateFontFace(fontObj);
        cssStyle.push(css2);
      }
      const filename = style === "normal" ? `${axesLower}.css` : `${axesLower}-${style}.css`;
      const cssPath = path.join(opts.dir, filename);
      const css = cssStyle.join("\n\n");
      await fs.writeFile(cssPath, css);
      if (axes === "wght")
        indexCSS = css;
      if (!indexCSS && axes === "opsz")
        indexCSS = css;
    }
  }
  await fs.writeFile(path.join(opts.dir, "index.css"), indexCSS);
};

const build = async (id, opts) => {
  const font = APIv2[id];
  const fontVariable = APIVariable[id];
  const fontLicense = APILicense[id];
  await fs.mkdir(opts.dir, { recursive: true });
  let changed = false;
  try {
    await fs.access(path.join(opts.dir, "metadata.json"));
    const metadata = JSON.parse(
      await fs.readFile(path.join(opts.dir, "metadata.json"), "utf8")
    );
    changed = metadata.lastModified !== font.lastModified;
  } catch {
    changed = true;
  }
  if (changed || opts.force) {
    try {
      const pkgJsonPath = path.join(opts.dir, "package.json");
      const tempPath = path.join(
        opts.tmpDir,
        opts.isVariable ? `${id}-variable.json` : `${id}.json`
      );
      await fs.access(pkgJsonPath);
      await fs.copy(pkgJsonPath, tempPath);
      await fs.emptyDir(opts.dir);
      await fs.copy(tempPath, pkgJsonPath);
      await fs.remove(tempPath);
    } catch {
    }
    await download(id, opts);
    if (opts.isVariable) {
      await packagerVariable(id, opts);
    } else {
      await packagerV1(id, opts);
      await packagerV2(id, opts);
    }
    const metadata = {
      id,
      family: font.family,
      subsets: font.subsets,
      weights: font.weights,
      styles: font.styles,
      defSubset: font.defSubset,
      variable: fontVariable ? fontVariable.axes : false,
      lastModified: font.lastModified,
      version: font.version,
      category: font.category,
      license: {
        type: licenseShort(fontLicense.license.type) ?? fontLicense.license.type,
        url: fontLicense.license.url,
        attribution: fontLicense.original
      },
      source: "https://github.com/google/fonts",
      type: "google"
    };
    await fs.mkdir(path.join(opts.dir, "scss"));
    await fs.writeFile(
      path.join(opts.dir, "scss/metadata.scss"),
      sassMetadata(metadata, font.unicodeRange, opts.isVariable)
    );
    await fs.writeFile(path.join(opts.dir, "scss/mixins.scss"), sassMixins);
    await fs.writeFile(
      path.join(opts.dir, "README.md"),
      readme(metadata, opts.isVariable)
    );
    await fs.writeFile(
      path.join(opts.dir, "unicode.json"),
      stringify(font.unicodeRange)
    );
    await fs.writeFile(path.join(opts.dir, "CHANGELOG.md"), changelog);
    await generateLicense(id, fontLicense.license.type, opts);
    await fs.writeFile(
      path.join(opts.dir, "metadata.json"),
      stringify(metadata)
    );
    await packageJson(metadata, opts);
  }
};

const queue = new PQueue({ concurrency: 3 });
queue.on("error", (error) => {
  throw new Error(error);
});
queue.on("idle", async () => {
  consola.success(
    `All ${Object.keys(APIv2).length} Google Fonts have been processed.`
  );
  consola.success(
    `${Object.keys(APIVariable).length} variable fonts have been processed.`
  );
});
const buildPackage = async (id, opts) => {
  consola.info(`Downloading ${id}`);
  const optsNew = {
    ...opts,
    dir: path.join(opts.dir, "google", id)
  };
  await build(id, optsNew);
  consola.success(`Finished processing ${id}`);
};
const buildVariablePackage = async (id, opts) => {
  consola.info(`Downloading ${id} ${colors.bold(colors.yellow("[VARIABLE]"))}`);
  const optsNew = {
    ...opts,
    isVariable: true,
    dir: path.join(opts.dir, "variable", id)
  };
  await build(id, optsNew);
  consola.success(
    `Finished processing ${id} ${colors.bold(colors.yellow("[VARIABLE]"))}`
  );
};
const buildIconPackage = async (id, opts) => {
  consola.info(`Downloading ${id} ${colors.bold(colors.green("[ICON]"))}`);
  const optsNew = {
    ...opts,
    dir: path.join(opts.dir, "icons", id)
  };
  await build(id, optsNew);
  consola.success(
    `Finished processing ${id} ${colors.bold(colors.green("[ICON]"))}`
  );
};
const buildVariableIconPackage = async (id, opts) => {
  consola.info(
    `Downloading ${id} ${colors.bold(colors.yellow("[VARIABLE ICON]"))}`
  );
  const optsNew = {
    ...opts,
    isVariable: true,
    dir: path.join(opts.dir, "variable-icons", id)
  };
  await build(id, optsNew);
  consola.success(
    `Finished processing ${id} ${colors.bold(colors.yellow("[VARIABLE ICON]"))}`
  );
};
const testIds = [
  "akshar",
  "alegreya",
  "archivo",
  "ballet",
  "fraunces",
  "noto-sans-jp",
  "recursive",
  "roboto-flex"
];
const processGoogle = async (opts, fonts) => {
  const outDir = path.resolve(process.cwd(), "fonts");
  await fs.ensureDir(outDir);
  const tmpDir = path.join(outDir, "fontsource_temp_packages");
  await fs.ensureDir(tmpDir);
  let fontIds = fonts;
  if (opts.test) {
    fontIds = testIds;
  } else if (!fonts || fonts.length === 0) {
    fontIds = Object.keys(APIv2);
  }
  for (const id of fontIds) {
    const buildOpts = {
      dir: outDir,
      tmpDir,
      isVariable: false,
      isIcon: false,
      force: opts.force ?? false
    };
    try {
      if (id in APIv2) {
        queue.add(() => buildPackage(id, buildOpts));
        if (APIVariable[id]) {
          queue.add(() => buildVariablePackage(id, buildOpts));
        }
      } else {
        consola.warn(`Skipping ${id} as it is not a Google Font.`);
      }
    } catch (error) {
      throw new Error(`${id} experienced an error. ${error}`);
    }
  }
  if (!fonts || fonts.length === 0) {
    fontIds = Object.keys(APIIconStatic);
  }
  for (const id of fontIds) {
    const buildOpts = {
      dir: outDir,
      tmpDir,
      isVariable: false,
      isIcon: true,
      force: opts.force ?? false
    };
    try {
      if (id in APIIconStatic) {
        queue.add(() => buildIconPackage(id, buildOpts));
        if (APIIconVariable[id]) {
          queue.add(() => buildVariableIconPackage(id, buildOpts));
        }
      } else {
        consola.warn(`Skipping ${id} as it is not a Google Font.`);
      }
    } catch (error) {
      throw new Error(`${id} experienced an error. ${error}`);
    }
  }
  await queue.onIdle();
  await fs.rm(tmpDir, { recursive: true });
};

const cli = cac("fontsource");
cli.command("fetch [key]", "Fetch parsing metadata for all fonts").option("-f, --force", "Force parse all metadata").action(async (key, options) => {
  try {
    const finalKey = key ?? process.env.GOOGLE_API_KEY;
    if (!finalKey) {
      throw new Error("No API key provided.");
    }
    if (options.force) {
      consola.info(
        `Parsing all metadata... ${colors.bold(colors.red("[FORCE]"))}`
      );
    } else {
      consola.info("Parsing all metadata...");
    }
    await Promise.all([fetchAPI(finalKey), fetchVariable()]);
    await parsev1(options.force, true);
    await parsev2(options.force, true);
    await generateAxis();
    await parseVariable(true);
    await parseIcons(options.force);
    await parseLicenses();
  } catch (error) {
    consola.error(error);
  }
});
cli.command("build [...fonts]", "Build font packages").option("-f, --force", "Force rebuild all packages").option("-t, --test", "Build test fonts only").action(async (fonts, options) => {
  try {
    consola.info(
      `Building packages... ${options.force ? colors.bold(colors.red("[FORCE]")) : ""}`
    );
    await processGoogle(options, fonts);
  } catch (error) {
    consola.error(error);
  }
});
cli.command("create").action(async () => {
  try {
    await create();
  } catch (error) {
    consola.error(error);
  }
});
cli.command("create-verify").action(async () => {
  try {
    await verify();
  } catch (error) {
    consola.error(error);
  }
});
cli.help();
cli.version(version);
cli.parse();
